import { ReactNode, ReactElement } from 'react';
import PropTypes from 'prop-types';

declare enum StatusEnum {
    ERROR = "error",
    IDLE = "idle",
    LOADING = "loading",
    SUCCESS = "success"
}

declare class PromiseType<T = unknown> extends Promise<T> {
    cancel?: () => void;
}
declare type OnErrorFn = ((err: Error) => void | Promise<void>) | undefined;
declare type OnSuccessFn<T = unknown> = ((result: T) => void | Promise<void>) | undefined;
declare type OnCancelFn = (() => void | Promise<void>) | undefined;
declare type NeedPollingType<T = unknown> = boolean | ((data: T) => boolean);
/**
 * @typedef {Object} UseDataLoaderConfig
 * @property {Function} [onSuccess] callback when a request success
 * @property {Function} [onError] callback when a error is occured, this will override the onError specified on the Provider if any
 * @property {*} [initialData] initial data if no one is present in the cache before the request
 * @property {number} [pollingInterval] relaunch the request after the last success
 * @property {boolean} [enabled=true] launch request automatically (default true)
 * @property {boolean} [keepPreviousData=true] do we need to keep the previous data after reload (default true)
 * @property {NeedPollingType} [needPolling=true] When pollingInterval is set you can set a set a custom callback to know if polling is enabled
 */
interface UseDataLoaderConfig<T = unknown> {
    enabled?: boolean;
    initialData?: T;
    keepPreviousData?: boolean;
    onError?: OnErrorFn;
    onSuccess?: OnSuccessFn;
    pollingInterval?: number;
    /**
     * Max time before data from previous success is considered as outdated (in millisecond)
     */
    maxDataLifetime?: number;
    needPolling?: NeedPollingType;
}
/**
 * @typedef {Object} UseDataLoaderResult
 * @property {boolean} isIdle true if the hook in initial state
 * @property {boolean} isLoading true if the request is launched
 * @property {boolean} isSuccess true if the request success
 * @property {boolean} isError true if the request throw an error
 * @property {boolean} isPolling true if the request if enabled is true, pollingInterval is defined and the status is isLoading or isSuccess
 * @property {*} previousData if keepPreviousData is true it return the last data fetched
 * @property {*} data initialData if no data is fetched or not present in the cache otherwise return the data fetched
 * @property {string} error the error occured during the request
 * @property {Function} reload reload the data
 */
interface UseDataLoaderResult<T = unknown> {
    data?: T;
    error?: Error;
    isError: boolean;
    isIdle: boolean;
    isLoading: boolean;
    isPolling: boolean;
    isSuccess: boolean;
    previousData?: T;
    reload: () => Promise<void>;
}
/**
 * Params send to the method
 */
declare type UsePaginatedDataLoaderMethodParams = {
    page: number;
    perPage: number;
};
declare type UsePaginatedDataLoaderConfig<T = unknown> = {
    enabled?: boolean;
    initialData?: T;
    keepPreviousData?: boolean;
    onError?: OnErrorFn;
    onSuccess?: OnSuccessFn;
    pollingInterval?: number;
    /**
     * Max time before data from previous success is considered as outdated (in millisecond)
     */
    maxDataLifetime?: number;
    needPolling?: NeedPollingType;
    initialPage?: number;
    perPage?: number;
};
declare type UsePaginatedDataLoaderResult<T = unknown> = {
    pageData?: T;
    data?: Record<number, T | undefined>;
    error?: Error;
    isError: boolean;
    isIdle: boolean;
    isLoading: boolean;
    isPolling: boolean;
    isSuccess: boolean;
    reload: () => Promise<void>;
    goToPage: (page: number) => void;
    goToNextPage: () => void;
    goToPreviousPage: () => void;
    page: number;
};

declare type DataLoaderConstructorArgs<T = unknown> = {
    key: string;
    method: () => PromiseType<T>;
    pollingInterval?: number;
    needPolling?: NeedPollingType<T>;
    maxDataLifetime?: number;
    keepPreviousData?: boolean;
};
declare class DataLoader<T = unknown> {
    static started: number;
    static maxConcurrent: number;
    static cachedData: Record<string, unknown>;
    key: string;
    status: StatusEnum;
    pollingInterval?: number;
    needPolling: NeedPollingType<T>;
    maxDataLifetime?: number;
    isDataOutdated: boolean;
    method: () => PromiseType<T>;
    private cancelMethod?;
    private canceled;
    keepPreviousData?: boolean;
    private errorListeners;
    private successListeners;
    private cancelListeners;
    private observerListeners;
    error?: Error;
    private dataOutdatedTimeout?;
    timeout?: number;
    private destroyed;
    constructor(args: DataLoaderConstructorArgs<T>);
    getData(): T | undefined;
    private notifyChanges;
    load: (force?: boolean) => Promise<void>;
    private launch;
    cancel: () => Promise<void>;
    addOnSuccessListener(fn: OnSuccessFn<T>): void;
    addOnErrorListener(fn: OnErrorFn): void;
    addOnCancelListener(fn: OnCancelFn): void;
    removeOnSuccessListener(fn: OnSuccessFn<T>): void;
    removeOnErrorListener(fn: OnErrorFn): void;
    removeOnCancelListener(fn: OnCancelFn): void;
    addObserver(fn: (args: DataLoader<T>) => void): void;
    removeObserver(fn: (args: DataLoader<T>) => void): void;
    clearData(): void;
    destroy(): Promise<void>;
    getObserversCount(): number;
    setPollingInterval(newPollingInterval?: number): void;
    setNeedPolling(needPolling: NeedPollingType<T>): void;
}

declare type CachedData = Record<string, unknown>;
declare type Reloads = Record<string, () => Promise<void | unknown>>;
declare type UseDataLoaderInitializerArgs<T = unknown> = {
    status?: StatusEnum;
    method: () => PromiseType<T>;
    pollingInterval?: number;
    keepPreviousData?: boolean;
    /**
     * Max time before data from previous success is considered as outdated (in millisecond)
     */
    maxDataLifetime?: number;
    needPolling?: NeedPollingType<T>;
};
declare type GetCachedDataFn = {
    (): CachedData;
    (key?: string): unknown | undefined;
};
declare type GetReloadsFn = {
    (): Reloads;
    (key?: string): (() => Promise<void | unknown>) | undefined;
};
interface IDataLoaderContext {
    addRequest: (key: string, args: UseDataLoaderInitializerArgs) => DataLoader;
    getOrAddRequest: (key: string, args: UseDataLoaderInitializerArgs) => DataLoader;
    cacheKeyPrefix?: string;
    onError?: (error: Error) => void | Promise<void>;
    clearAllCachedData: () => void;
    clearCachedData: (key: string) => void;
    getCachedData: GetCachedDataFn;
    getReloads: GetReloadsFn;
    getRequest: (key: string) => DataLoader;
    reload: (key?: string) => Promise<void>;
    reloadAll: () => Promise<void>;
}
declare const DataLoaderProvider: {
    ({ children, cacheKeyPrefix, onError, maxConcurrentRequests, }: {
        children: ReactNode;
        cacheKeyPrefix: string;
        onError: OnErrorFn;
        maxConcurrentRequests?: number | undefined;
    }): ReactElement;
    propTypes: {
        cacheKeyPrefix: PropTypes.Requireable<string>;
        children: PropTypes.Validator<string | number | boolean | {} | PropTypes.ReactElementLike | PropTypes.ReactNodeArray>;
        maxConcurrentRequests: PropTypes.Requireable<number>;
        onError: PropTypes.Requireable<(...args: any[]) => any>;
    };
    defaultProps: {
        cacheKeyPrefix: undefined;
        maxConcurrentRequests: number;
        onError: undefined;
    };
};
declare const useDataLoaderContext: () => IDataLoaderContext;

/**
 * @param {string} key key to save the data fetched in a local cache
 * @param {() => PromiseType} method a method that return a promise
 * @param {useDataLoaderConfig} config hook configuration
 * @returns {useDataLoaderResult} hook result containing data, request state, and method to reload the data
 */
declare const useDataLoader: <T>(fetchKey: string, method: () => PromiseType<T>, { enabled, initialData, keepPreviousData, onError, onSuccess, pollingInterval, maxDataLifetime, needPolling, }?: UseDataLoaderConfig<T>) => UseDataLoaderResult<T>;

/**
 * @param {string} baseFetchKey base key used to cache data. Hook append -page-X to that key for each page you load
 * @param {() => PromiseType} method a method that return a promise
 * @param {useDataLoaderConfig} config hook configuration
 * @returns {useDataLoaderResult} hook result containing data, request state, and method to reload the data
 */
declare const usePaginatedDataLoader: <T>(baseFetchKey: string, method: (params: UsePaginatedDataLoaderMethodParams) => PromiseType<T>, { enabled, initialData, keepPreviousData, onError, onSuccess, pollingInterval, maxDataLifetime, needPolling, initialPage, perPage, }?: UsePaginatedDataLoaderConfig<T>) => UsePaginatedDataLoaderResult<T>;

export { DataLoaderProvider, useDataLoader, useDataLoaderContext, usePaginatedDataLoader };
