import PropTypes from 'prop-types';
import { useRef, useCallback, useMemo, useContext, createContext, useState, useEffect, useLayoutEffect } from 'react';
import { jsx } from 'react/jsx-runtime';

let StatusEnum;

(function (StatusEnum) {
  StatusEnum["ERROR"] = "error";
  StatusEnum["IDLE"] = "idle";
  StatusEnum["LOADING"] = "loading";
  StatusEnum["SUCCESS"] = "success";
})(StatusEnum || (StatusEnum = {}));

const KEY_IS_NOT_STRING_ERROR = 'Key should be a string';
const DEFAULT_MAX_CONCURRENT_REQUESTS = 20;

class DataLoader {
  static started = 0;
  static maxConcurrent = DEFAULT_MAX_CONCURRENT_REQUESTS;
  static cachedData = {};
  needPolling = true;
  isDataOutdated = false;
  canceled = false;
  errorListeners = [];
  successListeners = [];
  cancelListeners = [];
  observerListeners = [];
  destroyed = false;

  constructor(args) {
    this.key = args.key;
    this.status = StatusEnum.IDLE;
    this.method = args.method;
    this.pollingInterval = args?.pollingInterval;
    this.keepPreviousData = args?.keepPreviousData;
    this.maxDataLifetime = args.maxDataLifetime;
    this.needPolling = args.needPolling ?? true;
    this.notifyChanges();
  }

  getData() {
    return DataLoader.cachedData[this.key] ?? undefined;
  }

  notifyChanges() {
    this.observerListeners.forEach(observerListener => observerListener(this));
  }

  load = async (force = false) => {
    if (force || this.status === StatusEnum.IDLE || this.status === StatusEnum.ERROR || this.status === StatusEnum.SUCCESS && this.isDataOutdated) {
      if (DataLoader.started < DataLoader.maxConcurrent) {
        // Because we want to launch the request directly without waiting the return
        if (this.timeout) {
          // Prevent multiple call at the same time
          clearTimeout(this.timeout);
        }

        await this.launch();
      } else {
        await new Promise(resolve => {
          setTimeout(resolve);
        }).then(() => this.load(force));
      }
    }
  };
  launch = async () => {
    try {
      if (this.status !== StatusEnum.LOADING) {
        this.canceled = false;
        this.status = StatusEnum.LOADING;
        this.notifyChanges();
      }

      DataLoader.started += 1;
      const promise = this.method();
      this.cancelMethod = promise.cancel;
      const result = await promise.then(res => res);
      this.status = StatusEnum.SUCCESS;
      this.error = undefined;

      if (!this.canceled) {
        DataLoader.cachedData[this.key] = result;
        await Promise.all(this.successListeners.map(listener => listener?.(result)));
        this.isDataOutdated = false;

        if (this.dataOutdatedTimeout) {
          clearTimeout(this.dataOutdatedTimeout);
          this.dataOutdatedTimeout = undefined;
        }

        if (this.maxDataLifetime) {
          this.dataOutdatedTimeout = setTimeout(() => {
            this.isDataOutdated = true;
            this.notifyChanges();
          }, this.maxDataLifetime);
        } else {
          this.isDataOutdated = true;
        }
      }

      this.notifyChanges();
    } catch (err) {
      this.status = StatusEnum.ERROR;
      this.error = err;
      this.notifyChanges();

      if (!this.canceled) {
        await Promise.all(this.errorListeners.map(listener => listener?.(err)));
      }
    }

    DataLoader.started -= 1;

    if (this.pollingInterval && !this.destroyed && (typeof this.needPolling === 'function' ? this.needPolling(DataLoader.cachedData[this.key]) : this.needPolling)) {
      this.timeout = setTimeout( // eslint-disable-next-line @typescript-eslint/no-misused-promises
      this.launch, this.pollingInterval);
    }
  };
  cancel = async () => {
    this.canceled = true;
    this.cancelMethod?.();
    await Promise.all(this.cancelListeners.map(listener => listener?.()));
  };

  addOnSuccessListener(fn) {
    if (!this.successListeners.includes(fn)) {
      this.successListeners.push(fn);
    }
  }

  addOnErrorListener(fn) {
    if (!this.errorListeners.includes(fn)) {
      this.errorListeners.push(fn);
    }
  }

  addOnCancelListener(fn) {
    if (!this.cancelListeners.includes(fn)) {
      this.cancelListeners.push(fn);
    }
  }

  removeOnSuccessListener(fn) {
    const index = this.successListeners.indexOf(fn);

    if (index > -1) {
      this.successListeners.splice(index, 1);
    }
  }

  removeOnErrorListener(fn) {
    const index = this.errorListeners.indexOf(fn);

    if (index > -1) {
      this.errorListeners.splice(index, 1);
    }
  }

  removeOnCancelListener(fn) {
    const index = this.cancelListeners.indexOf(fn);

    if (index > -1) {
      this.cancelListeners.splice(index, 1);
    }
  }

  addObserver(fn) {
    this.observerListeners.push(fn);
  }

  removeObserver(fn) {
    const index = this.observerListeners.indexOf(fn);

    if (index > -1) {
      this.observerListeners.splice(index, 1);
    }
  }

  clearData() {
    DataLoader.cachedData[this.key] = undefined;
    this.notifyChanges();
  }

  async destroy() {
    DataLoader.cachedData[this.key] = undefined;
    await this.cancel?.();

    if (this.timeout) {
      clearTimeout(this.timeout);
    }

    this.destroyed = true;
  }

  getObserversCount() {
    return this.observerListeners.length;
  }

  setPollingInterval(newPollingInterval) {
    this.pollingInterval = newPollingInterval;
  }

  setNeedPolling(needPolling) {
    this.needPolling = needPolling;
  }

}

// @ts-expect-error we force the context to undefined, should be corrected with default values
const DataLoaderContext = /*#__PURE__*/createContext(undefined);

const DataLoaderProvider = ({
  children,
  cacheKeyPrefix,
  onError,
  maxConcurrentRequests
}) => {
  const requestsRef = useRef({});
  const computeKey = useCallback(key => `${cacheKeyPrefix ? `${cacheKeyPrefix}-` : ''}${key}`, [cacheKeyPrefix]);
  const getRequest = useCallback(key => requestsRef.current[computeKey(key)], [computeKey]);
  const addRequest = useCallback((key, args) => {
    if (DataLoader.maxConcurrent !== maxConcurrentRequests) {
      DataLoader.maxConcurrent = maxConcurrentRequests;
    }

    if (key && typeof key === 'string') {
      const newRequest = new DataLoader({ ...args,
        key: computeKey(key)
      });
      requestsRef.current[newRequest.key] = newRequest;
      return newRequest;
    }

    throw new Error(KEY_IS_NOT_STRING_ERROR);
  }, [computeKey, maxConcurrentRequests]);
  const getOrAddRequest = useCallback((key, args) => {
    const requestFound = getRequest(key);

    if (!requestFound) {
      return addRequest(key, args);
    }

    return requestFound;
  }, [addRequest, getRequest]);
  const clearCachedData = useCallback(key => {
    if (typeof key === 'string') {
      if (requestsRef.current[computeKey(key)]) {
        requestsRef.current[computeKey(key)].clearData();
      }
    } else throw new Error(KEY_IS_NOT_STRING_ERROR);
  }, [computeKey]);
  const clearAllCachedData = useCallback(() => {
    Object.values(requestsRef.current).forEach(request => {
      request.clearData();
    });
  }, []);
  const reload = useCallback(async key => {
    if (key && typeof key === 'string') {
      await getRequest(key)?.load(true);
    } else throw new Error(KEY_IS_NOT_STRING_ERROR);
  }, [getRequest]);
  const reloadAll = useCallback(async () => {
    await Promise.all(Object.values(requestsRef.current).map(request => request.load(true)));
  }, []);
  const getCachedData = useCallback(key => {
    if (key) {
      return getRequest(key)?.getData();
    }

    return Object.values(requestsRef.current).reduce((acc, request) => ({ ...acc,
      [request.key]: request.getData()
    }), {});
  }, [getRequest]);
  const getReloads = useCallback(key => {
    if (key) {
      return getRequest(key) ? () => getRequest(key).load(true) : undefined;
    }

    return Object.entries(requestsRef.current).reduce((acc, [requestKey, {
      load
    }]) => ({ ...acc,
      [requestKey]: () => load(true)
    }), {});
  }, [getRequest]);
  const value = useMemo(() => ({
    addRequest,
    cacheKeyPrefix,
    clearAllCachedData,
    clearCachedData,
    getCachedData,
    getOrAddRequest,
    getReloads,
    getRequest,
    onError,
    reload,
    reloadAll
  }), [addRequest, cacheKeyPrefix, clearAllCachedData, clearCachedData, getCachedData, getOrAddRequest, getRequest, getReloads, onError, reload, reloadAll]);
  return /*#__PURE__*/jsx(DataLoaderContext.Provider, {
    value: value,
    children: children
  });
};

DataLoaderProvider.propTypes = {
  cacheKeyPrefix: PropTypes.string,
  children: PropTypes.node.isRequired,
  maxConcurrentRequests: PropTypes.number,
  onError: PropTypes.func
};
DataLoaderProvider.defaultProps = {
  cacheKeyPrefix: undefined,
  maxConcurrentRequests: DEFAULT_MAX_CONCURRENT_REQUESTS,
  onError: undefined
};
const useDataLoaderContext = () => useContext(DataLoaderContext);

/**
 * @param {string} key key to save the data fetched in a local cache
 * @param {() => PromiseType} method a method that return a promise
 * @param {useDataLoaderConfig} config hook configuration
 * @returns {useDataLoaderResult} hook result containing data, request state, and method to reload the data
 */
const useDataLoader = (fetchKey, method, {
  enabled = true,
  initialData,
  keepPreviousData = true,
  onError,
  onSuccess,
  pollingInterval,
  maxDataLifetime,
  needPolling
} = {}) => {
  const isMountedRef = useRef(false);
  const isFetchingRef = useRef(false);
  const unsubscribeRequestRef = useRef();
  const previousDataRef = useRef();
  const {
    getOrAddRequest,
    onError: onErrorProvider
  } = useDataLoaderContext();
  const [, forceReload] = useState(0);
  const subscribeFn = useCallback(() => {
    forceReload(x => x + 1);
  }, []);
  const request = useMemo(() => {
    if (unsubscribeRequestRef.current) {
      unsubscribeRequestRef.current();
    }

    const newRequest = getOrAddRequest(fetchKey, {
      keepPreviousData,
      maxDataLifetime,
      method,
      needPolling,
      pollingInterval
    });

    unsubscribeRequestRef.current = () => newRequest.removeObserver(subscribeFn);

    newRequest.addObserver(subscribeFn);
    return newRequest;
  }, [fetchKey, getOrAddRequest, maxDataLifetime, method, needPolling, pollingInterval, keepPreviousData, subscribeFn]);
  useEffect(() => {
    if (request.method !== method) {
      request.method = method;
    }

    request.addOnErrorListener(onError ?? onErrorProvider);
    request.addOnSuccessListener(onSuccess);
    return () => {
      request.removeOnErrorListener(onError ?? onErrorProvider);
      request.removeOnSuccessListener(onSuccess);
    };
  }, [onSuccess, onError, onErrorProvider, method, request]);
  const cancelMethodRef = useRef(request?.cancel);
  const isLoading = useMemo(() => enabled && request.status === StatusEnum.IDLE || request.status === StatusEnum.LOADING, [request.status, enabled]);
  const isIdle = useMemo(() => request.status === StatusEnum.IDLE, [request.status]);
  const isSuccess = useMemo(() => request.status === StatusEnum.SUCCESS, [request.status]);
  const isError = useMemo(() => request.status === StatusEnum.ERROR, [request.status]);
  const isPolling = useMemo(() => !!(enabled && pollingInterval && (isSuccess || isLoading)), [isSuccess, isLoading, enabled, pollingInterval]);
  useEffect(() => {
    if (enabled) {
      // launch should never throw
      // eslint-disable-next-line @typescript-eslint/no-floating-promises
      request.load();
    }
  }, [request, enabled, isIdle]);
  useEffect(() => {
    if (pollingInterval !== request.pollingInterval) {
      request.setPollingInterval(pollingInterval);
    }
  }, [pollingInterval, request]);
  useEffect(() => {
    if (needPolling !== request.needPolling) {
      request.setNeedPolling(needPolling ?? true);
    }
  }, [needPolling, request]);
  useEffect(() => {
    isFetchingRef.current = isLoading || isPolling;
  }, [isLoading, isPolling]);
  useLayoutEffect(() => {
    cancelMethodRef.current = request?.cancel;
  }, [request?.cancel]);
  useEffect(() => {
    isMountedRef.current = true;
    return () => {
      isMountedRef.current = false;

      if (isFetchingRef.current && cancelMethodRef.current) {
        cancelMethodRef.current();
      }

      unsubscribeRequestRef.current?.();
    };
  }, []);
  useEffect(() => () => {
    if (request.getData() !== previousDataRef.current && keepPreviousData) {
      previousDataRef.current = request.getData();
    }
  });
  return {
    data: request.getData() ?? initialData,
    error: request?.error,
    isError,
    isIdle,
    isLoading,
    isPolling,
    isSuccess,
    previousData: previousDataRef.current,
    reload: () => request.load(true)
  };
};

/**
 * @param {string} baseFetchKey base key used to cache data. Hook append -page-X to that key for each page you load
 * @param {() => PromiseType} method a method that return a promise
 * @param {useDataLoaderConfig} config hook configuration
 * @returns {useDataLoaderResult} hook result containing data, request state, and method to reload the data
 */

const usePaginatedDataLoader = (baseFetchKey, method, {
  enabled = true,
  initialData,
  keepPreviousData = true,
  onError,
  onSuccess,
  pollingInterval,
  maxDataLifetime,
  needPolling,
  initialPage,
  perPage = 1
} = {}) => {
  if (typeof baseFetchKey !== 'string') {
    throw new Error(KEY_IS_NOT_STRING_ERROR);
  }

  const [data, setData] = useState({});
  const [page, setPage] = useState(initialPage ?? 1);
  const pageMethod = useCallback(() => method({
    page,
    perPage
  }), [method, page, perPage]);
  const {
    data: pageData,
    isError,
    isIdle,
    isLoading,
    isPolling,
    isSuccess,
    reload,
    error
  } = useDataLoader(`${baseFetchKey}-page-${page}`, pageMethod, {
    enabled,
    initialData,
    keepPreviousData,
    maxDataLifetime,
    needPolling,
    onError,
    onSuccess,
    pollingInterval
  });
  const goToNextPage = useCallback(() => {
    setPage(current => current + 1);
  }, []);
  const goToPreviousPage = useCallback(() => {
    setPage(current => current > 1 ? current - 1 : 1);
  }, []);
  const goToPage = useCallback(newPage => {
    setPage(newPage > 1 ? newPage : 1);
  }, []);
  useEffect(() => {
    setData(current => {
      if (pageData !== current[page]) {
        return { ...current,
          [page]: pageData
        };
      }

      return current;
    });
  }, [pageData, page]);
  useEffect(() => {
    setPage(1);
    setData({});
  }, [baseFetchKey]);
  return {
    data,
    error,
    goToNextPage,
    goToPage,
    goToPreviousPage,
    isError,
    isIdle,
    isLoading,
    isPolling,
    isSuccess,
    page,
    pageData,
    reload
  };
};

export { DataLoaderProvider, useDataLoader, useDataLoaderContext, usePaginatedDataLoader };
